跳到论文正文
W Papers独立论文阅读器 arXiv:1211.3642 论文正文 计算阅读时间 · 0%
引用这篇论文根据当前论文元数据生成,正式投稿前建议核对。

    
  
License: arXiv.org perpetual non-exclusive license
arXiv:1211.3642v2 [cs.DS] 18 Jan 2013

Simpler and Faster Lempel Ziv Factorization

Keisuke Goto    Hideo Bannai Affiliation: Department of Informatics, Kyushu University, Fukuoka 819-0395, Japan Email: {keisuke.gotou,bannai}@inf.kyushu-u.ac.jp
Abstract

We present a new, simple, and efficient approach for computing the Lempel-Ziv (LZ77) factorization of a string in linear time, based on suffix arrays. Computational experiments on various data sets show that our approach constantly outperforms the fastest previous algorithm LZ_OG (Ohlebusch and Gog 2011), and can be up to 2 to 3 times faster in the processing after obtaining the suffix array, while requiring the same or a little more space.

1 Introduction

The LZ77 factorization [1] of a string captures important properties concerning repeated occurrences of substrings in the string, and has obvious applications in the field of data compression, as well as being the key component to various efficient algorithms on strings [2, 3]. Consequently, many algorithms for its efficient calculation have been proposed. The LZ77 factorization of a string SS is a factorization S=f1⋯fnS=f_{1}\cdots f_{n} where each factor fkf_{k} is either (1) a single character if that character does not occur in f1⋯fk−1f_{1}\cdots f_{k-1}, or, (2) the longest prefix of the rest of the string which occurs at least twice in f1⋯fkf_{1}\cdots f_{k}.

A naïve algorithm that computes the longest common prefix with each of the O⁡(N)O(N) previous positions only requires O⁡(1)O(1) working space (excluding the output), but can take O⁡(N2)O(N^{2}) time, where NN is the length of the string. Using string indicies such as suffix trees [4] and on-line algorithms to construct them [5], the LZ factorization can be computed in an on-line manner in O⁡(N​log⁡|Σ|)O(N\log|\Sigma|) time and O⁡(N)O(N) space, where |Σ||\Sigma| is the size of the alphabet.

Most recent efficient linear time algorithms are off-line, running in O⁡(N)O(N) time for integer alphabets using O⁡(N)O(N) space (See Table 1). They first construct the suffix array [6] of the string, and compute an array called the Longest Previous Factor (LPF) array from which the LZ factorization can be easily computed [7, 8, 9, 10, 11]. Many algorithms of this family first compute the longest common prefix (LCP) array prior to the computation of the LPF array. However, the computation of the LCP array is also costly. The algorithm CI1 (COMPUTE_LPF) of [12], and the algorithm LZ_OG [10] cleverly avoids its computation and directly computes the LPF array.

An important observation here is that the LPF is actually more information than is required for the computation of the LZ factorization, i.e., if our objective is the LZ factorization, we only use a subset of the entries in the LPF . However, the above algorithms focus on computing the entire LPF array, perhaps since it is difficult to determine beforehand, which entries of LPF are actually required. Although some algorithms such as a variant of CPS1 [8] or CPS2 in [8] avoid computation of LPF, they either require the LCP array, or do not run in linear worst case time and are not as efficient. (See  [11] for a survey.)

In this paper, we propose a new approach to avoid the computation of LCP and LPF arrays altogether, by combining the ideas of the naïve algorithm with those of CI1 and LZ_OG, and still achieve worst case linear time. The resulting algorithm is surprisingly both simple and efficient.

Computational experiments on various data sets shows that our algorithm constantly outperforms LZ_OG [10], and can be up to 2 to 3 times faster in the processing after obtaining the suffix array, while requiring the same or a little more space.

Although our algorithm might be considered as a simple combination of ideas appearing in previous works, this paper is one of the first to propose, implement and evaluate this combination. We note that algorithms that avoid the computation of LCP and LPF based on similar ideas as in this paper were developed independently and almost simultaneously by Kempa and Puglisi [13] and Kärkkäinen et al. [14]. Since we did not have knowledge of their work until very recently, we have not made comparisons between them. The worst case time complexity of [13] is not independent of alphabet size, but is fast and space efficient. In the more recent manuscript [14], two new linear time algorithms which outperform all previous algorithms (including ours) in terms of time and space are proposed, asserting the potential of this approach.

Table 1: Fast Linear time LZ-Factorization Algorithms based on Suffix Arrays
algorithm worst case time SA LCP LPF, PrevOcc
CI1 [12] Θ⁡(N)\Theta(N) ✓ ✓
CI2 [12] Θ⁡(N)\Theta(N) ✓ ✓ ✓
CPS1 [8] Θ⁡(N)\Theta(N) ✓ ✓
CIS [7] Θ⁡(N)\Theta(N) ✓ ✓ ✓
CII [9] Θ⁡(N)\Theta(N) ✓ ✓ ✓
LZ_OG [10] Θ⁡(N)\Theta(N) ✓ ✓
this work, [14] Θ⁡(N)\Theta(N) ✓
Naive Θ⁡(N2)\Theta(N^{2})

2 Preliminaries

Let 𝒩\mathcal{N} be the set of non-negative integers. Let Σ\Sigma be a finite alphabet. An element of Σ∗\Sigma^{*} is called a string. The length of a string TT is denoted by |T||T|. The empty string ε\varepsilon is the string of length 0, namely, |ε|=0|\varepsilon|=0. Let Σ+=Σ∗−{ε}\Sigma^{+}=\Sigma^{*}-\{\varepsilon\}. For a string S=X​Y​ZS=XYZ, XX, YY and ZZ are called a prefix, substring, and suffix of TT, respectively. The set of prefixes of TT is denoted by 𝑝𝑟𝑒𝑓𝑖𝑥⁡(T)\mathit{prefix}(T). The longest common prefix of strings X,YX,Y, denoted 𝑙𝑐𝑝⁡(X,Y)\mathit{lcp}(X,Y), is the longest string in 𝑝𝑟𝑒𝑓𝑖𝑥⁡(X)∩𝑝𝑟𝑒𝑓𝑖𝑥⁡(Y)\mathit{prefix}(X)\cap\mathit{prefix}(Y).

The ii-th character of a string TT is denoted by T⁡[i]T[i] for 1≤i≤|T|1\leq i\leq|T|, and the substring of a string TT that begins at position ii and ends at position jj is denoted by T[i..j]T[i..j] for 1≤i≤j≤|T|1\leq i\leq j\leq|T|. For convenience, let T[i..j]=εT[i..j]=\varepsilon if j<ij<i, and T⁡[|T|+1]=$T[|T|+1]=\$ where $\$ is a special delimiter character that does not occur elsewhere in the string.

2.1 Suffix Arrays

The suffix array [6] 𝑆𝐴\mathit{SA} of any string TT is an array of length |T||T| such that for any 1≤i≤|T|1\leq i\leq|T|, 𝑆𝐴⁡[i]=j\mathit{SA}[i]=j indicates that T[j:|T|]T[j:|T|] is the ii-th lexicographically smallest suffix of TT. For convenience, assume that 𝑆𝐴⁡[0]=|T|+1\mathit{SA}[0]=|T|+1. The inverse array 𝑆𝐴−1\mathit{SA}^{-1} of 𝑆𝐴\mathit{SA} is an array of length |T||T| such that 𝑆𝐴−1​[𝑆𝐴​[i]]=i\mathit{SA}^{-1}[\mathit{SA}[i]]=i. As in [15], let Φ\Phi be an array of length |T||T| such that Φ⁡[𝑆𝐴⁡[1]]=|T|\Phi[\mathit{SA}[1]]=|T| and Φ⁡[𝑆𝐴⁡[i]]=𝑆𝐴⁡[i−1]\Phi[\mathit{SA}[i]]=\mathit{SA}[i-1] for 2≤i≤|T|2\leq i\leq|T|, i.e., for any suffix j=𝑆𝐴⁡[i]j=\mathit{SA}[i], Φ⁡[j]=𝑆𝐴⁡[i−1]\Phi[j]=\mathit{SA}[i-1] is the immediately preceding suffix in the suffix array. The suffix array 𝑆𝐴\mathit{SA} for any string of length |T||T| can be constructed in O⁡(|T|)O(|T|) time regardless of the alphabet size, assuming an integer alphabet (e.g. [16]). All our algorithms will assume that the 𝑆𝐴\mathit{SA} is already computed. Given 𝑆𝐴\mathit{SA}, arrays 𝑆𝐴−1\mathit{SA}^{-1} and Φ\Phi can easily be computed in linear time by a simple scan.

2.2 LZ Encodings

LZ encodings are dynamic dictionary based encodings with many variants. The variant we consider is also known as the s-factorization [17].

Definition 1 (LZ77-factorization)

The s-factorization of a string TT is the factorization T=f1⋯fnT=f_{1}\cdots f_{n} where each s-factor fk∈Σ+​(k=1,…,n)f_{k}\in\Sigma^{+}~(k=1,\ldots,n) is defined inductively as follows: f1=T⁡[1]f_{1}=T[1]. For k≥2k\geq 2: if T[|f1⋯fk−1|+1]=c∈ΣT[|f_{1}\cdots f_{k-1}|+1]=c\in\Sigma does not occur in f1⋯fk−1f_{1}\cdots f_{k-1}, then fk=cf_{k}=c. Otherwise, fkf_{k} is the longest prefix of fk⋯fnf_{k}\cdots f_{n} that occurs at least twice in f1⋯fkf_{1}\cdots f_{k}.

Note that each LZ factor can be represented in constant space, i.e., a pair of integers where the first and second elements respectively represent the length and position of a previous occurrence of the factor. If the factor is a new character and the length of its previous occurrence is 00, the second element will encode the new character instead of the position. For example the s-factorization of the string T=𝚊𝚋𝚊𝚊𝚋𝚊𝚋𝚊𝚋𝚊𝚊𝚊𝚊𝚊𝚋𝚋𝚊𝚋𝚊𝚋T=\mathtt{abaabababaaaaabbabab} is 𝚊\mathtt{a}, 𝚋\mathtt{b}, 𝚊\mathtt{a}, 𝚊𝚋𝚊\mathtt{aba}, 𝚋𝚊𝚋𝚊\mathtt{baba}, 𝚊𝚊𝚊𝚊\mathtt{aaaa}, 𝚋\mathtt{b}, 𝚋𝚊𝚋𝚊𝚋\mathtt{babab}. This can be represented as (0,𝚊)(0,\mathtt{a}), (0,𝚋)(0,\mathtt{b}), (1,1)(1,1), (3,1)(3,1), (4,5)(4,5), (4,10)(4,10), (1,2)(1,2), (5,5)(5,5).

We define two functions 𝐿𝑃𝐹\mathit{LPF} and 𝑃𝑟𝑒𝑣𝑂𝑐𝑐\mathit{PrevOcc} below. For any 1≤i≤N1\leq i\leq N, 𝐿𝑃𝐹⁡(i)\mathit{LPF}(i) is the longest length of longest common prefix between T[i:N]T[i:N] and T[j:N]T[j:N] for any 1≤j<i1\leq j<i, and 𝑃𝑟𝑒𝑣𝑂𝑐𝑐⁡(i)\mathit{PrevOcc}(i) is a position jj which achieves gives 𝐿𝑃𝐹⁡(i)\mathit{LPF}(i)11 1 There can be multiple choices of jj, but here, it suffices to fix one.. More precisely,

𝐿𝑃𝐹⁡(i)\displaystyle\mathit{LPF}(i) =\displaystyle= max({0}∪{𝑙𝑐𝑝(T[i:N],T[j:N])∣1≤j<i})\displaystyle\max(\{0\}\cup\{\mathit{lcp}(T[i:N],T[j:N])\mid 1\leq j<i\})
and
𝑃𝑟𝑒𝑣𝑂𝑐𝑐⁡(i)\displaystyle\mathit{PrevOcc}(i) =\displaystyle= {−1if ​𝐿𝑃𝐹​(i)=0jotherwise\displaystyle\begin{cases}-1&\mbox{if }\mathit{LPF}(i)=0\\ j&\mbox{otherwise}\end{cases}

where jj satisfies 1≤j<i1\leq j<i, and T[i:i+𝐿𝑃𝐹(i)−1]=T[j:j+𝐿𝑃𝐹(i)−1]T[i:i+\mathit{LPF}(i)-1]=T[j:j+\mathit{LPF}(i)-1]. Let pk=|f1⋯fk−1|+1p_{k}=|f_{1}\cdots f_{k-1}|+1. Then, fkf_{k} can be represented as a pair (𝐿𝑃𝐹⁡(pk),𝑃𝑟𝑒𝑣𝑂𝑐𝑐⁡(pk))(\mathit{LPF}(p_{k}),\mathit{PrevOcc}(p_{k})) if 𝐿𝑃𝐹⁡(pk)>0\mathit{LPF}(p_{k})>0, and (0,T⁡[pk])(0,T[p_{k}]) otherwise.

Most recent fast linear time algorithms for computing the LZ factorization calculate 𝐿𝑃𝐹\mathit{LPF} and 𝑃𝑟𝑒𝑣𝑂𝑐𝑐\mathit{PrevOcc} for all positions 1≤i≤N1\leq i\leq N of the text and store the values in an array, and then use these values as in Algorithm 1 to output the LZ factorization.

Algorithm 1 LZ Factorization from 𝐿𝑃𝐹\mathit{LPF} and 𝑃𝑟𝑒𝑣𝑂𝑐𝑐\mathit{PrevOcc} arrays
Input : String TT, 𝐿𝑃𝐹\mathit{LPF}, 𝑃𝑟𝑒𝑣𝑂𝑐𝑐\mathit{PrevOcc}
1 p←1p\leftarrow 1;
2 while p≤Np\leq N do
    3 if 𝐿𝑃𝐹⁡[p]=0\mathit{LPF}[p]=0 then Output : (1,T⁡[p])(1,T[p])
    4 ;
    5 else Output : (𝐿𝑃𝐹⁡[p],𝑃𝑟𝑒𝑣𝑂𝑐𝑐⁡[p])(\mathit{LPF}[p],\mathit{PrevOcc}[p])
    6 ;
    7 p←p+max⁡(1,𝐿𝑃𝐹⁡[p])p\leftarrow p+\max(1,\mathit{LPF}[p]);

3 Algorithm

We first describe the naïve algorithm for calculating the LZ factorization of a string, and analyze its time complexity The naïve algorithm does not compute all values of 𝐿𝑃𝐹\mathit{LPF} and 𝑃𝑟𝑒𝑣𝑂𝑐𝑐\mathit{PrevOcc} as explicit arrays, but only the values required to represent each factor. The procedure is shown in Algorithm 2. For a factor starting at position pp, the algorithm computes 𝐿𝑃𝐹⁡(p)\mathit{LPF}(p) and 𝑃𝑟𝑒𝑣𝑂𝑐𝑐⁡(p)\mathit{PrevOcc}(p) by simply looking at each of its p−1p-1 previous positions, and naively computes the longest common prefix (lcp) between each previous suffix and the suffix starting at position pp, and outputs the factor accordingly. At first glance, this algorithm looks like an O⁡(N3)O(N^{3}) time algorithm since there are 3 nested loops. However, the total time can be bounded by O⁡(N2)O(N^{2}), since the total length of the longest lcp’s found for each pp in the algorithm, i.e., the total length of the LZ factors found, is NN. More precisely, let the LZ factorization of string TT of length NN be f1⋯fnf_{1}\cdots f_{n}, and pk=|f1⋯fk−1|+1p_{k}=|f_{1}\cdots f_{k-1}|+1 as before. Then, the number of character comparisons executed in Line 2 of Algorithm 2 when calculating fkf_{k} is at most (pk−1)​|fk+1|(p_{k}-1)|f_{k}+1|, and the total can be bounded: ∑k=1n(pk−1)​|fk+1|≤N​∑k=1n|fk+1|=O⁡(N2)\sum_{k=1}^{n}(p_{k}-1)|f_{k}+1|\leq N\sum_{k=1}^{n}|f_{k}+1|=O(N^{2}). An important observation here is that if we can somehow reduce the number of previous candidate positions for naïvely computing lcp’s (i.e. the choice of jj in Line 2 of Algorithm 2) from O⁡(N)O(N) to O⁡(1)O(1) positions, this would result in a O⁡(N)O(N) time algorithm. This very simple observation is the first key to the linear running times of our new algorithms.

Algorithm 2 Naïve Algorithm for Calculating LZ factorization
Input : String TT
1 p←1p\leftarrow 1;
2 while p≤|T|p\leq|T| do
    3 𝐿𝑃𝐹←0\mathit{LPF}\leftarrow 0;
    4 for j←1,…,p−1j\leftarrow 1,\ldots,p-1 do
       5 l←0l\leftarrow 0;
       6 while T⁡[j+l]=T⁡[p+l]T[j+l]=T[p+l] do l←l+1l\leftarrow l+1; // l←𝑙𝑐𝑝(T[j:N],T[p:N])l\leftarrow\mathit{lcp}(T[j:N],T[p:N])
       7 ;
       8 if l>𝐿𝑃𝐹l>\mathit{LPF} then 𝐿𝑃𝐹←l\mathit{LPF}\leftarrow l; 𝑃𝑟𝑒𝑣𝑂𝑐𝑐←j\mathit{PrevOcc}\leftarrow j;
       9 ;
    10 if 𝐿𝑃𝐹>0\mathit{LPF}>0 then Output : (𝐿𝑃𝐹,𝑃𝑟𝑒𝑣𝑂𝑐𝑐)(\mathit{LPF},\mathit{PrevOcc})
    11 ;
    12 else Output : (0,T⁡[p])(0,T[p])
    13 ;
    14 p←p+max⁡(1,𝐿𝑃𝐹)p\leftarrow p+\max(1,\mathit{LPF});

To accomplish this, our algorithm utilizes yet another simple but key observation made in [12]. Since suffixes in the suffix arrays are lexicographically sorted, if we fix a suffix 𝑆𝐴⁡[i]\mathit{SA}[i] in the suffix array, we know that suffixes appearing closer in the suffix array will have longer longest common prefixes with suffix 𝑆𝐴⁡[i]\mathit{SA}[i].

For any position 1≤i≤N1\leq i\leq N of the suffix array, let

𝑃𝑆𝑉𝑙𝑒𝑥​[i]\displaystyle\mathit{PSV_{\mathit{lex}}}[i] =\displaystyle= max⁡({0}∪{1≤j<i∣𝑆𝐴⁡[j]<𝑆𝐴⁡[i]})\displaystyle\max(\{0\}\cup\{1\leq j<i\mid\mathit{SA}[j]<\mathit{SA}[i]\})
𝑁𝑆𝑉𝑙𝑒𝑥​[i]\displaystyle\mathit{NSV_{\mathit{lex}}}[i] =\displaystyle= min⁡({0}∪{N≥j>i∣𝑆𝐴⁡[j]<𝑆𝐴⁡[i]})\displaystyle\min(\{0\}\cup\{N\geq j>i\mid\mathit{SA}[j]<\mathit{SA}[i]\})

i.e., for the suffix starting at text position 𝑆𝐴⁡[i]\mathit{SA}[i], the values 𝑃𝑆𝑉𝑙𝑒𝑥​[i]\mathit{PSV_{\mathit{lex}}}[i] and 𝑁𝑆𝑉𝑙𝑒𝑥​[i]\mathit{NSV_{\mathit{lex}}}[i] represent the lexicographic rank of the suffixes that start before it in the string and are lexicographically closest (previous and next) to it, or 00 if such a suffix does not exist. From the above arguments, we have that for any text position 1≤p≤N1\leq p\leq N,

𝐿𝑃𝐹(p)=max(𝑙𝑐𝑝(T[𝑆𝐴[𝑃𝑆𝑉𝑙𝑒𝑥[𝑆𝐴−1[p]]]:N],T[p:N]),\displaystyle\mathit{LPF}(p)=\max(\mathit{lcp}(T[\mathit{SA}[\mathit{PSV_{\mathit{lex}}}[\mathit{SA}^{-1}[p]]]:N],T[p:N]),
𝑙𝑐𝑝(T[𝑆𝐴[𝑁𝑆𝑉𝑙𝑒𝑥[𝑆𝐴−1[p]]]:N],T[p:N])).\displaystyle\mathit{lcp}(T[\mathit{SA}[\mathit{NSV_{\mathit{lex}}}[\mathit{SA}^{-1}[p]]]:N],T[p:N])).

The above observation or its variant has been used as the basis for calculating 𝐿𝑃𝐹⁡(i)\mathit{LPF}(i) for all 1≤i≤N1\leq i\leq N in linear time in practically all previous linear time algorithms for LZ factorization based on the suffix array. In [10], they consider (implicitly) the arrays in text order rather than lexicographic order. In this case,

𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[𝑆𝐴​[i]]\displaystyle\mathit{PSV_{\mathit{text}}}[\mathit{SA}[i]] =\displaystyle= 𝑆𝐴​[𝑃𝑆𝑉𝑙𝑒𝑥​[i]]\displaystyle\mathit{SA}[\mathit{PSV_{\mathit{lex}}}[i]]
𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[𝑆𝐴​[i]]\displaystyle\mathit{NSV_{\mathit{text}}}[\mathit{SA}[i]] =\displaystyle= 𝑆𝐴​[𝑁𝑆𝑉𝑙𝑒𝑥​[i]]\displaystyle\mathit{SA}[\mathit{NSV_{\mathit{lex}}}[i]]

and therefore

𝐿𝑃𝐹(p)=max(𝑙𝑐𝑝(T[𝑃𝑆𝑉𝑡𝑒𝑥𝑡[p]]:N],T[p:N]),𝑙𝑐𝑝(T[𝑁𝑆𝑉𝑡𝑒𝑥𝑡[p]]:N],T[p:N])).\displaystyle\mathit{LPF}(p)=\max(\mathit{lcp}(T[\mathit{PSV_{\mathit{text}}}[p]]:N],T[p:N]),\mathit{lcp}(T[\mathit{NSV_{\mathit{text}}}[p]]:N],T[p:N])).

While [12] and [10] utilize this observation to compute all entries of 𝐿𝑃𝐹\mathit{LPF} in linear time, we utilize it in a slightly different way as mentioned previously, and use it to reduce the candidate positions for calculating 𝑃𝑟𝑒𝑣𝑂𝑐𝑐⁡(i)\mathit{PrevOcc}(i) (i.e. the choice of jj in Algorithm 2) to only 2 positions. The key idea of our approach is in the combination of the above observation with the amortized analysis of the naïve algorithm, suggesting that we can defer the computation of the values of 𝐿𝑃𝐹\mathit{LPF} until we actually require them for the LZ factorization and still achieve linear worst case time. If 𝑃𝑆𝑉𝑙𝑒𝑥​[i]\mathit{PSV_{\mathit{lex}}}[i] and 𝑁𝑆𝑉𝑙𝑒𝑥​[i]\mathit{NSV_{\mathit{lex}}}[i] (or 𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[i]\mathit{PSV_{\mathit{text}}}[i] and 𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[i]\mathit{NSV_{\mathit{text}}}[i]) are known for all 1≤i≤N1\leq i\leq N, the linear running time of the algorithm follows from the previous arguments. The basic structure of our algorithm is shown in Algorithm 3 when using 𝑃𝑆𝑉𝑙𝑒𝑥\mathit{PSV_{\mathit{lex}}} and 𝑁𝑆𝑉𝑙𝑒𝑥\mathit{NSV_{\mathit{lex}}}. Note that it is easy to replace them with 𝑃𝑆𝑉𝑡𝑒𝑥𝑡\mathit{PSV_{\mathit{text}}} and 𝑁𝑆𝑉𝑡𝑒𝑥𝑡\mathit{NSV_{\mathit{text}}}, and in such case, 𝑆𝐴\mathit{SA} and 𝑆𝐴−1\mathit{SA}^{-1} are not necessary once we have 𝑃𝑆𝑉𝑡𝑒𝑥𝑡\mathit{PSV_{\mathit{text}}} and 𝑁𝑆𝑉𝑡𝑒𝑥𝑡\mathit{NSV_{\mathit{text}}}.

What remains is how to compute 𝑃𝑆𝑉𝑙𝑒𝑥​[i]\mathit{PSV_{\mathit{lex}}}[i] and 𝑁𝑆𝑉𝑙𝑒𝑥​[i]\mathit{NSV_{\mathit{lex}}}[i], or, 𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[i]\mathit{PSV_{\mathit{text}}}[i] and 𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[i]\mathit{NSV_{\mathit{text}}}[i] for all 1≤i≤N1\leq i\leq N. This can be done in several ways. We consider 3 variations.

The first is a computation of 𝑃𝑆𝑉𝑙𝑒𝑥​[i]\mathit{PSV_{\mathit{lex}}}[i], 𝑁𝑆𝑉𝑙𝑒𝑥​[i]\mathit{NSV_{\mathit{lex}}}[i] using a simple linear time scan of the suffix array with the help of a stack. The procedure is shown in Algorithm 4. This variant requires the text, and the arrays 𝑆𝐴\mathit{SA}, 𝑆𝐴−1\mathit{SA}^{-1}, 𝑃𝑆𝑉𝑙𝑒𝑥\mathit{PSV_{\mathit{lex}}}, 𝑁𝑆𝑉𝑙𝑒𝑥\mathit{NSV_{\mathit{lex}}} and a stack. The total space complexity is 17​N+4​Sm​a​x17N+4S_{max} bytes assuming that an integer occupies 4 bytes, where Sm​a​xS_{max} is the maximum size of the stack during the execution of the algorithm and can be Θ⁡(n)\Theta(n) in the worstcase. We will call this variant BGS.

The other two is a process called peak elimination, which is very briefly described in [12] for lexicographic order (Shown in Algorithms 5 and 6), and in [10] for text order (Shown in Algorithms 7 and 8). In peak elimination, each suffix ii and its lexicographically preceding suffix jj (𝑆𝐴−1​[j]+1=𝑆𝐴−1​[i]\mathit{SA}^{-1}[j]+1=\mathit{SA}^{-1}[i]) is examined in some order of ii (lexicographic or text order). For simplicity, we only briefly explain the approach for text order. If i>ji>j, this means that 𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[i]=j\mathit{PSV_{\mathit{text}}}[i]=j and if i<ji<j, 𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[j]=i\mathit{NSV_{\mathit{text}}}[j]=i. When both values of 𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[i]\mathit{PSV_{\mathit{text}}}[i] and 𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[i]\mathit{NSV_{\mathit{text}}}[i] are determined, ii is identified as a peak. Given a peak ii, it is possible to eliminate it, and determine the value of either 𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[i]]\mathit{NSV_{\mathit{text}}}[\mathit{PSV_{\mathit{text}}}[i]] (which will be 𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[i]\mathit{NSV_{\mathit{text}}}[i] if 𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[i]>𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[i]\mathit{PSV_{\mathit{text}}}[i]>\mathit{NSV_{\mathit{text}}}[i]) or 𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[i]]\mathit{PSV_{\mathit{text}}}[\mathit{NSV_{\mathit{text}}}[i]] (which will be 𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[i]\mathit{PSV_{\mathit{text}}}[i] if if 𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[i]<𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[i]\mathit{PSV_{\mathit{text}}}[i]<\mathit{NSV_{\mathit{text}}}[i])), and this process is repeated. The algorithm runs in linear time since each position can be eliminated only once. The procedure for lexicographic order is a bit simpler since the lexicographic order of calculation implies that 𝑃𝑆𝑉𝑙𝑒𝑥​[i]\mathit{PSV_{\mathit{lex}}}[i] will always be determined before 𝑁𝑆𝑉𝑙𝑒𝑥​[i]\mathit{NSV_{\mathit{lex}}}[i].

The algorithm of [10] actually computes the arrays 𝐿𝑃𝐹\mathit{LPF} and 𝑃𝑟𝑒𝑣𝑂𝑐𝑐\mathit{PrevOcc} directly without computing 𝑃𝑆𝑉𝑡𝑒𝑥𝑡\mathit{PSV_{\mathit{text}}} and 𝑁𝑆𝑉𝑡𝑒𝑥𝑡\mathit{NSV_{\mathit{text}}}. The algorithm we show is actually a simplification, deferring the computation of 𝐿𝑃𝐹\mathit{LPF} and 𝑃𝑟𝑒𝑣𝑂𝑐𝑐\mathit{PrevOcc}, computing 𝑃𝑆𝑉𝑡𝑒𝑥𝑡\mathit{PSV_{\mathit{text}}} and 𝑁𝑆𝑉𝑡𝑒𝑥𝑡\mathit{NSV_{\mathit{text}}} instead.

For lexicographic order, we need the text and the arrays 𝑆𝐴\mathit{SA}, 𝑆𝐴−1\mathit{SA}^{-1}, 𝑃𝑆𝑉𝑙𝑒𝑥\mathit{PSV_{\mathit{lex}}}, 𝑁𝑆𝑉𝑙𝑒𝑥\mathit{NSV_{\mathit{lex}}} and no stack, giving an algorithm with 17​N17N bytes of working space. We will call this variant BGL. For text order, although the Φ\Phi array is introduced instead of the 𝑆𝐴−1\mathit{SA}^{-1} array, the suffix array is not required after its computation. Therefore, by reusing the space of 𝑆𝐴\mathit{SA} for 𝑃𝑆𝑉𝑡𝑒𝑥𝑡\mathit{PSV_{\mathit{text}}}, the total space complexity can be reduced to 13​N13N bytes. We will call this variant BGT. Note that although p​e​a​k​E​l​i​ml​e​xpeakElim_{lex} and p​e​a​k​E​l​i​mt​e​x​tpeakElim_{text} are shown as recursive functions for simplicity, they are tail recursive and thus can be optimized as loops and will not require extra space on the call stack.

3.1 Interleaving 𝑃𝑆𝑉\mathit{PSV} and 𝑁𝑆𝑉\mathit{NSV}

Since accesses to 𝑃𝑆𝑉\mathit{PSV} and 𝑁𝑆𝑉\mathit{NSV} occur at the same or close indices, it is possible to improve the memory locality of accesses by interleaving the values of 𝑃𝑆𝑉\mathit{PSV} and 𝑁𝑆𝑉\mathit{NSV}, maintaining them in a single array as follows. Let 𝑃𝑁𝑆𝑉\mathit{PNSV} be an array of length 2​N2N, and for each position 1≤i≤2​N1\leq i\leq 2N, 𝑃𝑁𝑆𝑉⁡[i]=𝑃𝑆𝑉⁡[j]\mathit{PNSV}[i]=\mathit{PSV}[j] if imod2≡0i\mod 2\equiv 0, 𝑁𝑆𝑉⁡[j]\mathit{NSV}[j] otherwise, where j=⌊i/2⌋j=\lfloor i/2\rfloor. Naturally, for any 1≤i≤N1\leq i\leq N, 𝑃𝑆𝑉\mathit{PSV} and 𝑁𝑆𝑉\mathit{NSV} can be accessed as 𝑃𝑆𝑉⁡[i]=𝑃𝑁𝑆𝑉⁡[2​i]\mathit{PSV}[i]=\mathit{PNSV}[2i] and 𝑁𝑆𝑉⁡[i]=𝑃𝑁𝑆𝑉⁡[2​i+1]\mathit{NSV}[i]=\mathit{PNSV}[2i+1]. This interleaving can be done for both lexicographic order and text order. We will call the variants of our algorithms that incorporate this optimization, iBGS, iBGL, iBGT.

Algorithm 3 Basic Structure of our Algorithms.
Input : String TT
1 Calculate 𝑃𝑆𝑉𝑙𝑒𝑥​[i]\mathit{PSV_{\mathit{lex}}}[i] and 𝑁𝑆𝑉𝑙𝑒𝑥​[i]\mathit{NSV_{\mathit{lex}}}[i] for all i=1​…​Ni=1...N;
2 p←1p\leftarrow 1;
3 while p≤Np\leq N do
    4 𝐿𝑃𝐹←0\mathit{LPF}\leftarrow 0;
    5 for j∈{𝑆𝐴⁡[𝑃𝑆𝑉𝑙𝑒𝑥​[𝑆𝐴−1​[p]]],𝑆𝐴⁡[𝑁𝑆𝑉𝑙𝑒𝑥​[𝑆𝐴−1​[p]]]}j\in\{\mathit{SA}[\mathit{PSV_{\mathit{lex}}}[\mathit{SA}^{-1}[p]]],\mathit{SA}[\mathit{NSV_{\mathit{lex}}}[\mathit{SA}^{-1}[p]]]\} do
       6 l←0l\leftarrow 0;
       7 while T⁡[j+l]=T⁡[p+l]T[j+l]=T[p+l] do l←l+1l\leftarrow l+1; // l←𝑙𝑐𝑝(T[j:N],T[p:N])l\leftarrow\mathit{lcp}(T[j:N],T[p:N])
       8 ;
       9 if l>𝐿𝑃𝐹l>\mathit{LPF} then 𝐿𝑃𝐹←l\mathit{LPF}\leftarrow l; 𝑃𝑟𝑒𝑣𝑂𝑐𝑐←j\mathit{PrevOcc}\leftarrow j;
       10 ;
    11 if 𝐿𝑃𝐹>0\mathit{LPF}>0 then Output : (𝐿𝑃𝐹,𝑃𝑟𝑒𝑣𝑂𝑐𝑐)(\mathit{LPF},\mathit{PrevOcc})
    12 ;
    13 else Output : (0,T⁡[p])(0,T[p])
    14 ;
    15 p←p+max⁡(1,𝐿𝑃𝐹)p\leftarrow p+\max(1,\mathit{LPF});
Algorithm 4 Calculating 𝑃𝑆𝑉𝑙𝑒𝑥\mathit{PSV_{\mathit{lex}}} and 𝑁𝑆𝑉𝑙𝑒𝑥\mathit{NSV_{\mathit{lex}}} from 𝑆𝐴\mathit{SA}
Input : Suffix array 𝑆𝐴\mathit{SA}
Output : 𝑃𝑆𝑉𝑙𝑒𝑥\mathit{PSV_{\mathit{lex}}}, 𝑁𝑆𝑉𝑙𝑒𝑥\mathit{NSV_{\mathit{lex}}}
1 Let SS be an empty stack;
2 for i←1i\leftarrow 1 to NN do
    3 x←𝑆𝐴⁡[i]x\leftarrow\mathit{SA}[i];
    4 while (𝐧𝐨𝐭S.𝑒𝑚𝑝𝑡𝑦())𝐚𝐧𝐝(𝑆𝐴[S.𝑡𝑜𝑝()]>x)(\mathbf{not}~S.\mathit{empty}())~\mathbf{and}~(\mathit{SA}[S.\mathit{top}()]>x) do
       5 𝑁𝑆𝑉𝑙𝑒𝑥[S.𝑡𝑜𝑝()]←i\mathit{NSV_{\mathit{lex}}}[S.\mathit{top}()]\leftarrow i; S.𝑝𝑜𝑝⁡()S.\mathit{pop}() ;
    6 𝑃𝑆𝑉𝑙𝑒𝑥​[i]←\mathit{PSV_{\mathit{lex}}}[i]\leftarrow if S.e​m​p​t​y​()S.empty() then 00 ;
    7 else S.t​o​p​()S.top();
    8 ;
    9 S.p​u​s​h​(i)S.push(i);
10 while 𝐧𝐨𝐭​S.𝑒𝑚𝑝𝑡𝑦⁡()\mathbf{not}~S.\mathit{empty}() do
    11 𝑁𝑆𝑉𝑙𝑒𝑥[S.top()]←0\mathit{NSV_{\mathit{lex}}}[S.top()]\leftarrow 0; S.p​o​p​()S.pop() ;
Algorithm 5 Calculating 𝑃𝑆𝑉𝑙𝑒𝑥\mathit{PSV_{\mathit{lex}}} and 𝑁𝑆𝑉𝑙𝑒𝑥\mathit{NSV_{\mathit{lex}}} from 𝑆𝐴\mathit{SA} by Peak Elimination.
Input : Suffix array 𝑆𝐴\mathit{SA}
1 for i←1i\leftarrow 1 to NN do 𝑁𝑆𝑉𝑙𝑒𝑥​[i]←0\mathit{NSV_{\mathit{lex}}}[i]\leftarrow 0;
2 ;
3 𝑃𝑆𝑉𝑙𝑒𝑥​[1]←0\mathit{PSV_{\mathit{lex}}}[1]\leftarrow 0;
4 for i←2i\leftarrow 2 to NN do p​e​a​k​E​l​i​ml​e​x​(i−1,i)peakElim_{lex}(i-1,i);
5 ;
Algorithm 6 Peak Elimination p​e​a​k​E​l​i​ml​e​x​(j,i)peakElim_{lex}(j,i) in Lexicographic Order.
if j=0j=0 or 𝑆𝐴⁡[j]<𝑆𝐴⁡[i]\mathit{SA}[j]<\mathit{SA}[i] then
    1 𝑃𝑆𝑉𝑙𝑒𝑥​[i]←j\mathit{PSV_{\mathit{lex}}}[i]\leftarrow j;
2 else // j≥1j\geq 1 and 𝑆𝐴⁡[j]>𝑆𝐴⁡[i]\mathit{SA}[j]>\mathit{SA}[i]
    3 𝑁𝑆𝑉𝑙𝑒𝑥​[j]←i\mathit{NSV_{\mathit{lex}}}[j]\leftarrow i;
    4 p​e​a​k​E​l​i​ml​e​x​(𝑃𝑆𝑉𝑙𝑒𝑥​[j],i)peakElim_{lex}(\mathit{PSV_{\mathit{lex}}}[j],i) ; // jj was peak.
Algorithm 7 Calculating 𝑃𝑆𝑉𝑡𝑒𝑥𝑡\mathit{PSV_{\mathit{text}}} and 𝑁𝑆𝑉𝑡𝑒𝑥𝑡\mathit{NSV_{\mathit{text}}} from 𝑆𝐴\mathit{SA} using Φ\Phi.
Input : Suffix array 𝑆𝐴\mathit{SA}
1 Φ⁡[𝑆𝐴⁡[1]]←N\Phi[\mathit{SA}[1]]\leftarrow N;
2 for i←2i\leftarrow 2 to NN do Φ⁡[𝑆𝐴⁡[i]]←𝑆𝐴⁡[i−1]\Phi[\mathit{SA}[i]]\leftarrow\mathit{SA}[i-1];
3 ;
4 for i←1i\leftarrow 1 to NN do
    5 𝑃𝑆𝑉𝑡𝑒𝑥𝑡[i]←⊥\mathit{PSV_{\mathit{text}}}[i]\leftarrow\bot; 𝑁𝑆𝑉𝑡𝑒𝑥𝑡[i]←⊥\mathit{NSV_{\mathit{text}}}[i]\leftarrow\bot;
6 for i←1i\leftarrow 1 to NN do p​e​a​k​E​l​i​mt​e​x​t​(Φ⁡[i],i)peakElim_{text}(\Phi[i],i);
7 ;
Algorithm 8 Peak Elimination p​e​a​k​E​l​i​mt​e​x​t​(j,i)peakElim_{text}(j,i)
if j<ij<i then
    1 𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[i]←j\mathit{PSV_{\mathit{text}}}[i]\leftarrow j;
    2 if 𝑁𝑆𝑉𝑡𝑒𝑥𝑡[i]≠⊥\mathit{NSV_{\mathit{text}}}[i]\neq\bot then p​e​a​k​E​l​i​mt​e​x​t​(j,𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[i])peakElim_{text}(j,\mathit{NSV_{\mathit{text}}}[i]) ; // ii was peak.
    3 ;
4 else // j>ij>i
    5 𝑁𝑆𝑉𝑡𝑒𝑥𝑡​[j]←i\mathit{NSV_{\mathit{text}}}[j]\leftarrow i;
    6 if 𝑃𝑆𝑉𝑡𝑒𝑥𝑡[j]≠⊥\mathit{PSV_{\mathit{text}}}[j]\neq\bot then p​e​a​k​E​l​i​mt​e​x​t​(𝑃𝑆𝑉𝑡𝑒𝑥𝑡​[j],i)peakElim_{text}(\mathit{PSV_{\mathit{text}}}[j],i) ; // jj was peak.
    7 ;

4 Computational Experiments

We implement and compare our algorithms with LZ_OG since it has been shown to be the most time efficient in the experiments of [10]. We also implement a variant LZ_iOG which incorporates the interleaving optimization for 𝐿𝑃𝐹\mathit{LPF} and 𝑃𝑟𝑒𝑣𝑂𝑐𝑐\mathit{PrevOcc} arrays. We have made the source codes publicly available at http://code.google.com/p/lzbg/.

All computations were conducted on a Mac Xserve (Early 2009) with 2 x 2.93GHz Quad Core Xeon processors and 24GB Memory, only utilizing a single process/thread at once. The programs were compiled using the GNU C++ compiler (g++) 4.2.1 with the -fast option for optimization. The running times are measured in seconds, starting from after the suffix array is built, and the average of 10 runs is reported.

We use the data of http://www.cas.mcmaster.ca/~bill/strings/, used in previous work. Table 2 shows running times of the algorithms, as well as some statistics of the dataset. The running times of the fastest algorithm for each data is shown in bold. The fastest running times for the variant that uses only 13​N13N bytes is prefixed with ‘⊳\triangleright’.

The results show that all the variants of our algorithms constantly outperform LZ_OG and even LZ_iOG for all data tested, and in some cases can be up to 2 to 3 times faster. We can see that iBGS is fastest when the data is not extremely repetitive, and the average length of the factor is not so large, while iBGT is fastest for such highly repetitive data. iBGT is also the fastest when we restrict our attention to the algorithms that use only 13​N13N bytes of work space.

Table 2: Running times (seconds) of algorithms and various statistics for the data set of http://www.cas.mcmaster.ca/~bill/strings/.
LZ_OG LZ_iOG BGS iBGS BGL iBGL BGT iBGT
File Name 13​N13N 13​N13N 17​N+4​Sm​a​x17N+4S_{max} 17​N+4​Sm​a​x17N+4S_{max} 17​N17N 17​N17N 13​N13N 13​N13N |Σ||\Sigma| Text Size NN # of LZ factors Average Length of Factor Sm​a​xS_{max}
E.coli 0.64 0.58 0.26 0.23 0.33 0.29 0.45 ⊳\triangleright 0.37 4 4638690 432791 10.72 36
bible 0.37 0.34 0.20 0.19 0.25 0.22 0.27 ⊳\triangleright 0.24 63 4047392 337558 11.99 42
chr19.dna4 10.05 9.25 4.40 4.00 5.33 4.71 7.64 ⊳\triangleright 6.54 4 63811651 4411679 14.46 58
chr22.dna4 5.37 4.91 2.27 2.06 2.77 2.44 4.09 ⊳\triangleright 3.45 4 34553758 2554184 13.53 43
fib_s2178309 0.06 0.06 0.05 0.06 0.06 0.05 0.05 ⊳\triangleright 0.05 2 2178309 31 70268.00 16
fib_s3524578 0.11 0.11 0.10 0.10 0.10 0.10 0.10 ⊳\triangleright 0.09 2 3524578 32 110143.00 16
fib_s5702887 0.18 0.18 0.15 0.16 0.16 0.15 0.15 ⊳\triangleright 0.14 2 5702887 33 172815.00 17
fib_s9227465 0.30 0.30 0.26 0.27 0.27 0.26 0.26 ⊳\triangleright 0.24 2 9227465 34 271396.00 17
fib_s14930352 0.50 0.49 0.43 0.44 0.44 0.43 0.42 ⊳\triangleright 0.39 2 14930352 35 426581.00 18
fss9 0.09 0.08 0.08 0.08 0.08 0.08 0.07 ⊳\triangleright 0.07 2 2851443 40 71286.10 22
fss10 0.40 0.39 0.36 0.37 0.36 0.35 0.34 ⊳\triangleright 0.32 2 12078908 44 274521.00 24
howto 4.20 3.91 2.30 2.15 2.79 2.51 3.28 ⊳\triangleright 2.91 197 39422105 3063929 12.87 616
mozilla 5.30 4.95 3.19 3.13 3.91 3.65 4.31 ⊳\triangleright 3.86 256 51220480 6898100 7.43 3964
p1Mb 0.08 0.07 0.05 0.05 0.06 0.06 0.05 ⊳\triangleright 0.05 23 1048576 216146 4.85 38
p2Mb 0.23 0.21 0.11 0.12 0.15 0.15 0.17 ⊳\triangleright 0.14 23 2097152 406188 5.16 40
p4Mb 0.58 0.52 0.26 0.26 0.35 0.33 0.43 ⊳\triangleright 0.35 23 4194304 791583 5.30 42
p8Mb 1.27 1.15 0.55 0.55 0.73 0.70 0.94 ⊳\triangleright 0.78 23 8388608 1487419 5.64 898
p16Mb 2.70 2.43 1.18 1.16 1.52 1.46 2.08 ⊳\triangleright 1.74 23 16777216 2751022 6.10 898
p32Mb 5.58 5.02 2.47 2.44 3.14 3.03 4.43 ⊳\triangleright 3.74 24 33554432 5040051 6.66 898
rndA2_4Mb 0.49 0.45 0.20 0.18 0.24 0.20 0.35 ⊳\triangleright 0.28 2 4194304 201910 20.77 36
rndA2_8Mb 1.08 0.99 0.42 0.38 0.50 0.43 0.77 ⊳\triangleright 0.63 2 8388608 385232 21.78 37
rndA21_4Mb 0.64 0.58 0.28 0.28 0.38 0.37 0.47 ⊳\triangleright 0.37 21 4194304 970256 4.32 34
rndA21_8Mb 1.43 1.28 0.61 0.60 0.83 0.79 1.05 ⊳\triangleright 0.85 21 8388608 1835235 4.57 37
rndA255_4Mb 0.65 0.58 0.38 0.39 0.51 0.47 0.49 ⊳\triangleright 0.40 255 4194304 2005584 2.09 35
rndA255_8Mb 1.43 1.27 0.84 0.84 1.12 1.04 1.10 ⊳\triangleright 0.92 255 8388608 3817588 2.20 38

Acknowledgments: We thank Dr. Simon Gog for sharing his implementation of LZ_OG, and Dr. Simon Puglisi for sending us the manuscripts [13, 14].

References

  • [1] J. Ziv and A. Lempel, “A universal algorithm for sequential data compression,” IEEE Transactions on Information Theory, vol. IT-23, no. 3, pp. 337–343, 1977.
  • [2] R. Kolpakov and G. Kucherov, “Finding maximal repetitions in a word in linear time,” in Proc. FOCS 1999, 1999, pp. 596–604.
  • [3] J.-P. Duval, R. Kolpakov, G. Kucherov, T. Lecroq, and A. Lefebvre, “Linear-time computation of local periods,” Theoretical Computer Science, vol. 326, no. 1-3, pp. 229–240, 2004.
  • [4] P. Weiner, “Linear pattern-matching algorithms,” in Proc. of 14th IEEE Ann. Symp. on Switching and Automata Theory. Institute of Electrical Electronics Engineers, New York, 1973, pp. 1–11.
  • [5] E. Ukkonen, “On-line construction of suffix trees,” Algorithmica, vol. 14, no. 3, pp. 249–260, 1995.
  • [6] U. Manber and G. Myers, “Suffix arrays: A new method for on-line string searches,” SIAM J. Computing, vol. 22, no. 5, pp. 935–948, 1993.
  • [7] M. Crochemore, L. Ilie, and W. F. Smyth, “A simple algorithm for computing the Lempel Ziv factorization,” in Proc. DCC 2008, 2008, pp. 482–488.
  • [8] G. Chen, S. Puglisi, and W. Smyth, “Lempel-Ziv factorization using less time & space,” Mathematics in Computer Science, vol. 1, no. 4, pp. 605–623, 2008.
  • [9] M. Crochemore, L. Ilie, C. S. Iliopoulos, M. Kubica, W. Rytter, and T. Waleń, “LPF computation revisited,” in Proc. IWOCA 2009, 2009, pp. 158–169.
  • [10] E. Ohlebusch and S. Gog, “Lempel-Ziv factorization revisited,” in Proc. CPM’11, 2011, pp. 15–26.
  • [11] A. Al-Hafeedh, M. Crochemore, L. Ilie, J. Kopylov, W. Smyth, G. Tischler, and M. Yusufu, “A comparison of index-based Lempel-Ziv LZ77 factorization algorithms,” ACM Computing Surveys, in press.
  • [12] M. Crochemore and L. Ilie, “Computing longest previous factor in linear time and applications,” Information Processing Letters, vol. 106, no. 2, pp. 75–80, 2008.
  • [13] D. Kempa and S. J. Puglisi, “Lempel-Ziv factorization: Simple, fast, practical,” in Proc. ALENEX’13, 2013.
  • [14] J. Kärkkäinen, D. Kempa, and S. J. Puglisi, “Linear time Lempel-Ziv factorization: Simple, fast, small,” 2012, arXiv:1212.2952.
  • [15] J. Kärkkäinen, G. Manzini, and S. J. Puglisi, “Permuted longest-common-prefix array,” in CPM, 2009, pp. 181–192.
  • [16] J. Kärkkäinen and P. Sanders, “Simple linear work suffix array construction,” in Proc. ICALP 2003, 2003, pp. 943–955.
  • [17] M. Crochemore, “Linear searching for a square in a word,” Bulletin of the European Association of Theoretical Computer Science, vol. 24, pp. 66–72, 1984.